 # -*- coding: utf-8 -*-
import cPickle
import calendar
import datetime
from collections import defaultdict
from dateutil import relativedelta

from django.conf import settings
from django.db.models.fields import FieldDoesNotExist
from django.db.models.base import ModelBase
from django.db import models
from django.utils.functional import cached_property

from m3 import RelatedError

from educommon.utils.date import WEEKDAYS

from . import models as ioc_models
from . validators import BaseTimeTableValidator


class TimeTableBase(ModelBase):

    u"""
    Метакласс, занимающийся генерацией внешних ключей на модели.

    В т.ч. продуктовые (Специалист, Кабинет, Посетитель), и внешние ключи
    на модели расписания приема специалистов (внутри данного приложения)

    .. note::

        Метакласс для трех моделей общий, т.к. необходимо хранить состояние
        генерации классов (моделей) для организации взаимных внешних ключей
    """

    # cостоние генерации:
    # модель расписания сетки приемов
    TimeTableCrontab = None
    # модель приёма в сетке расписания (прием специалиста)
    TimeTableRecord = None

    @staticmethod
    def foreign_key(model_class, related_name=None):
        u"""Создание поля внешнего ключа на модель ``model_class``."""
        return models.ForeignKey(
            to=model_class,
            verbose_name=model_class._meta.verbose_name,
            related_name=related_name
        )

    @classmethod
    def __add_to_crontabbase(cls, model):
        u"""
        Создание внешних ключей на модель Специалиста и на Кабинет.

        (в модели Расписания сетки приёмов)
        """
        cls.TimeTableCrontab = model
        try:
            model._meta.get_field("specialist")
        except FieldDoesNotExist:
            model.add_to_class(
                "specialist", cls.foreign_key(
                    ioc_models.Specialist, 'timetablecrontab_set'
                )
            )

        try:
            model._meta.get_field("office")
        except FieldDoesNotExist:
            model.add_to_class(
                "office", cls.foreign_key(
                    ioc_models.Office, 'timetablecrontab_set')
            )

    @classmethod
    def __add_to_recordbase(cls, model):
        u"""
        Создание поля внешнего ключа на модель Расписания.

        (в модели Приёма в сетке расписания)
        """
        cls.TimeTableRecord = model

        if cls.TimeTableCrontab:
            try:
                model._meta.get_field("timetablecrontab")
            except FieldDoesNotExist:
                model.add_to_class(
                    "timetablecrontab", cls.foreign_key(
                        cls.TimeTableCrontab, 'timetablerecord_set'
                    )
                )

    @classmethod
    def __add_to_reception(cls, model):
        u"""
        Создание внешних ключей на модель Посетителя и на модель Приёма.

        (в модели-связке Приёма в сетке расписания и Посетителя)
        """
        try:
            model._meta.get_field("applicant")
        except FieldDoesNotExist:
            model.add_to_class(
                "applicant", cls.foreign_key(
                    ioc_models.Applicant, 'applicantreception_set'
                )
            )

        if cls.TimeTableRecord:
            try:
                model._meta.get_field("timetablerecord")
            except FieldDoesNotExist:
                model.add_to_class(
                    "timetablerecord", cls.foreign_key(
                        cls.TimeTableRecord, 'applicantreception_set'
                    )
                )
        # уникальность
        model._meta.unique_together = [('applicant', 'timetablerecord'), ]

    def __new__(cls, name, bases, attrs):
        model = super(TimeTableBase, cls).__new__(cls, name, bases, attrs)

        for b in bases:
            if b.__name__ == 'AbstractCrontab':
                cls.__add_to_crontabbase(model)

            elif b.__name__ == 'AbstractRecord':
                cls.__add_to_recordbase(model)

            elif b.__name__ == 'AbstractReception':
                cls.__add_to_reception(model)
        return model


class AbstractCrontab(models.Model):

    u"""Абстрактная модель сетки расписания приёма."""

    __metaclass__ = TimeTableBase

    validator_cls = None
    u"""
    Класс валидатора сохранения - наследник BaseTimeTableValidator
    """

    # выбор типа периодичности
    PERIOD_MONTH = 1
    PERIOD_WEEK = 2
    PERIOD_TYPES = [
        (PERIOD_MONTH, u'Ежемесячно'), (PERIOD_WEEK, u'Еженедельно')
    ]

    # макс.значения продолжительности периода
    MAX_PERIOD_LENGTH = {
        PERIOD_MONTH: 12,
        PERIOD_WEEK: 60
    }

    # варианты продолжительности приема в минутах
    # значение в строках - для комбобокса
    DURATION_TYPES = [
        (1, '10'), (2, '20'), (3, '30'), (4, '40'), (5, '50'),
        (6, '60'), (7, '70'), (8, '80'), (9, '90')
    ]
    # недель в месяце (и будет пять Понедельников, например)
    DEFAULT_WEEKS_COUNT = 5

    TIME_FORMAT = '%H:%M'

    # кортежи дней недели
    WEEKDAYS = WEEKDAYS

    # дата, с которой будет сформировано расписание приема
    date_begin = models.DateField(u'Дата начала')
    period_type = models.PositiveSmallIntegerField(
        verbose_name=u'Периодичность', choices=PERIOD_TYPES,
        default=PERIOD_MONTH,
    )
    # количество недель/месяцев
    period_length = models.PositiveSmallIntegerField(
        verbose_name=u'Продолжительность периода',
    )
    duration = models.PositiveSmallIntegerField(
        verbose_name=u'Продолжительность (мин.)', choices=DURATION_TYPES)

    u"""
    Хранит дамп настройки для формирования расписания в виде словаря
        {
            DAY_IDX: {'days': [2, 4, 5], 'begin': '17.45', 'end': '18.05'},
        }

    DAY_IDX - int, день из web_edu.core.fgos.schedule_template.models.WEEKDAYS
        например, Понедельник
    'days' - list, список int - номеров дней для периодичности PERIOD_MONTH
        например, [2, 4, 5] - 2-й, 4-й и 5-й Понедельник месяца
    'begin' - str, время начала
        например, 17:05
    'end' - str, время окончания
        например, 17:50

    Значение по ключу DAY_IDX может отсутствовать, это значит, что данный день
    не используется

    Поле - приватное, т.к. это дамп словаря, работать с ним нужно через
    свойство config
    """
    _config = models.TextField(
        verbose_name=u'Настройка расписания', default=u'', max_length=500)

    @property
    def config(self):
        # сPickle - чтобы не думать о типе данных
        return cPickle.loads(str(self._config))

    @config.setter
    def config(self, config):
        self._config = cPickle.dumps(config)

    @cached_property
    def _days_and_weeks(self):
        u"""
        Получение конфигурации месяца в виде словара.

        (зависимость номеров дней недели и номеров недель)
        :rtype: dict
        """
        # выбранные номера недель и дни недели в виде словаря
        days_and_weeks = defaultdict(list)
        week_range = range(1, self.DEFAULT_WEEKS_COUNT + 1)

        for day_idx, data in self.config.iteritems():
            if not data:
                continue

            _range = None
            if self.period_type == self.PERIOD_WEEK:
                _range = week_range
            else:
                if data['days']:
                    # дни были выбраны
                    _range = data['days']
            if _range:
                days_and_weeks[day_idx] = _range

        return days_and_weeks

    @cached_property
    def days_and_time(self):
        u"""
        Дни недели и датавремя (в формате datetime.datetime).

        :rtype: dict
        """
        time = lambda t: datetime.datetime.strptime(t, self.TIME_FORMAT)
        return dict((
            day_idx, (time(data['begin']), time(data['end']))
        ) for day_idx, data in self.config.iteritems() if data)

    def __init__(self, *args, **kwargs):
        assert issubclass(
            self.validator_cls, BaseTimeTableValidator
        ), u'Ожидается класс валидатора - наследник BaseTimeTableValidator'
        super(AbstractCrontab, self).__init__(*args, **kwargs)

    def iterdates(self):
        u"""
        Генератор дат по настройкам self.config.

        :warn: Даты отдаются в неотсортированном виде: сначала все ПН месяца,
        потом все ВТ месяца, ..., далее - след. месяц
        :rtype: datetime.date
        """

        # обход по периодичностям
        cc = calendar.Calendar(calendar.MONDAY)
        # шаг итерации по календарю
        delta = relativedelta.relativedelta(months=1)
        date = self.date_begin

        # дата последнего дня для формирования исходя из дня начала и
        # длины периода
        key = 'weeks' if self.period_type == self.PERIOD_WEEK else 'months'
        last_date = self.date_begin + relativedelta.relativedelta(
            **{key: self.period_length})

        # цикл по месяцам
        while date <= last_date:
            year, month = date.year, date.month
            month_start_date = datetime.date(year, month, 1)

            # цикл по календарю месяца (по столбцам)
            for day_idx, dates in enumerate(
                zip(*cc.monthdatescalendar(year, month))
            ):
                week_dates = [w_date for w_date in dates if (
                    # только текущего месяца
                    w_date >= month_start_date and
                    # позже даты старта
                    w_date >= self.date_begin and
                    # и не позже даты окончания
                    w_date < last_date

                )]
                for week_idx in self._days_and_weeks[day_idx]:
                    try:
                        yield week_dates[week_idx - 1]
                    except IndexError:
                        pass
            date += delta

    def get_timedelta_duration(self):
        u"""
        Получение продолжительности в виде datetime.timedelta.

        :rtype: datetime.timedelta
        """
        return datetime.timedelta(
            seconds=int(dict(self.DURATION_TYPES)[self.duration])*60)

    def applicant_reception_msg(self):
        u"""
        Проверка на наличие назначенных приемов Посетителей.

        у специалиста из timetablecrontab
        :returns: сообщение о наличии либо пустую строку
        :rtype: unicode
        """
        message = u''

        with_receptions = self.timetablerecord_set.filter(
            applicantreception_set__isnull=False
        )
        if with_receptions.exists():
            times = [begin.strftime('{0} {1}'.format(
                settings.DATE_FORMAT, self.TIME_FORMAT)
            ) for begin in with_receptions.values_list('begin', flat=True)]

            message = u'У сотрудника {0} назначены приемы: {1}'.format(
                self.specialist.fullname, u', '.join(times)
            )
        return message

    def clean(self):
        self.validator_cls(timetable=self)()

    def safe_delete(self):
        # нельзя удалять, раз посетителям назначены приемы
        has_child_msg = self.applicant_reception_msg()
        if has_child_msg:
            raise RelatedError(has_child_msg)
        else:
            self.timetablerecord_set.all().delete()
            return super(AbstractCrontab, self).safe_delete()

    class Meta:
        verbose_name = u'Расписание специалиста'
        abstract = True


class AbstractRecord(models.Model):

    u"""Абстрактная модель 'Прием' (запись сетки расписания)."""

    __metaclass__ = TimeTableBase

    # текст в ячейке, когда на прием не назначены посетители
    NO_RECEPTION = u'Нет приема'

    begin = models.DateTimeField(verbose_name=u'С')
    end = models.DateTimeField(verbose_name=u'По')

    class Meta:
        abstract = True
        verbose_name = u'Приём специалиста'
        verbose_name_plural = u'Приёмы специалиста'


class AbstractReception(models.Model):

    u"""
    Абстрактная модель 'Назначенный приём'.

    (связка посетителя с приёмом в сетке расписания).
    """

    __metaclass__ = TimeTableBase

    class Meta:
        abstract = True
        verbose_name = u'Назначенный приём'
        verbose_name_plural = u'Назначенные приёмы'
