# -*- coding: utf-8 -*-
import datetime
import abc

from django.db.transaction import commit_on_success

import objectpack
from m3 import actions as m3_actions

from educommon.objectpack.actions import BaseGridPack

from .. import models
from .. import providers
from .. api import accept_reception
from . ui import TimeTableWindow


class BaseReceptionPack(BaseGridPack):

    u"""Базовый пак расписания приёма специалиста."""

    title = u'Расписание приема специалиста'
    model = None
    u"""
    Модель записи приема специалиста. Наследник ``.base.TTRecordBase``

    .. code::

        model = models.TimeTableRecord

    """
    _is_primary_for_model = False

    window = TimeTableWindow

    column_param_name = providers.ColumnConstructor.column_param_name
    row_id_name = 'row_id'  # не используется в данном паке

    applicant_select_win_title_fmt = u"Посетители ({})"
    u"""
    Заголовок окна выбора посетителей для добавления на прием специалисту.
    В скобках будет указано действие: "добавить" или "изменить".


    .. code::
        applicant_select_win_title_fmt = u'Дети из заявлений ({})'
    """

    _msg_record_not_found = u'Запись расписания не найдена!'
    _msg_no_accepted_records = u'Нет назначенных приемов!'
    _msg_applicant_not_found = u'Посетитель не найден!'

    @abc.abstractproperty
    def specialist_pack(self):
        u"""
        Пак выбора специалистов.

        .. code::

            return ControllerCache.find_pack(SpecialistSelectPack)
        """

    @abc.abstractproperty
    def applicant_pack(self):
        u"""
        Пак посетителей для назначения и назначенных на прием.

        .. code::

            return ControllerCache.find_pack(ApplicantPack)
        """

    @abc.abstractproperty
    def schedule_print_pack(self):
        u"""
        Пак печати расписания.

        .. code::

            return ControllerCache.find_pack(SpecialistReceptionPrintPack)
        """

    def __init__(self):
        super(BaseReceptionPack, self).__init__()
        # экшн данных грида
        self.rows_action = RowsAction()
        # экшн показа окна выбора посетителя
        self.applicant_select_winaction = ApplicantSelectWinAction()
        # экшн назначения посетителя на прием
        self.applicantreception_save_action = ApplicantReceptionSaveAction()
        # экшн удаления назначения посетителя на прием
        self.delete_action = ApplicantReceptionDeleteAction()

        self.actions.extend([
            self.rows_action,
            self.applicant_select_winaction,
            self.applicantreception_save_action,
            self.delete_action
        ])

    def declare_context(self, action):
        result = super(BaseReceptionPack, self).declare_context(action)
        if action in (self.rows_action, self.grid_action):
            result.update(
                specialist_id={'type': 'int_or_none'},
                date={
                    'type': 'date', 'default': datetime.date.today()
                }
            )
        if action in (
            self.applicantreception_save_action,
            self.applicant_select_winaction,
            self.delete_action
        ):
            result[self.id_param_name] = {'type': 'int'}

        if action in (
            self.applicantreception_save_action,
            self.applicant_select_winaction,
        ):
            result['is_editing'] = {'type': 'boolean'}

        if action is self.applicantreception_save_action:
            result['id'] = {'type': 'int'}

        return result

    def get_rows_url(self):
        return self.rows_action.get_absolute_url()

    def get_print_url(self):
        return self.schedule_print_pack.report_window_action.get_absolute_url()

    def create_columns(self, request, context):
        u"""Возвращает заголовки колонок двумерного грида."""
        return providers.ColumnConstructor(context.date).create_columns()

    def set_grid_params(self, grid, request, context):
        params = super(BaseReceptionPack, self).set_grid_params(
            grid, request, context)
        params.update(
            add_url=self.applicant_select_winaction.get_absolute_url(),
            delete_url=self.delete_action.get_absolute_url()
        )
        return params

    def get_default_specialist(self, request, context):
        u"""
        Получение предвыбранного значения для поля 'Специалист'.

        .. code::

            # Пример для ЭШ
            # поиск по-умолчанию выбранного сотрудника (первого по алфавиту)
            # среди имеющих назначенные часы приема

            week = get_weekdays_for_date(_today)

            records = self.model.objects.filter(
                begin__lte=week[-1][1],
                end__gte=week[0][1],
                timetablecrontab__specialist__school=get_current_school(
                    request)
            ).order_by('timetablecrontab__specialist__person__fullname')

            # если есть право "просмотр своих" и нет права "просмотр всех"
            if self.has_perm(
                request, 'view_own'
            ) and not self.has_perm(request, 'view_all'):
                records = records.filter(
                    timetablecrontab__specialist__person__userprofile=(
                        request.user.get_profile())
                )

            if records.exists():
                specialist = records[0].timetablecrontab.specialist
            else:
                specialist = None
        """
        return None

    def get_window_params(self, request, context):
        u"""Передает параметры окну."""
        params = super(BaseReceptionPack, self).get_window_params(
            request, context)

        _today = datetime.date.today()

        params.update(
            read_only=not self.has_perm(request, 'edit'),
            date=_today,
            specialist_pack=self.specialist_pack.__class__,
            default_specialist=self.get_default_specialist(request, context),
            print_url=self.get_print_url()
        )
        return params

    def get_obj(self, context):
        u"""
        Получение объекта ``self.models``.

        по параметру self.id_param_name в контексте
        """
        return self.model.objects.get(id=getattr(context, self.id_param_name))

    def get_applicantreception_save_url(self):
        u"""Урл экшна назначения на прием / изменения приема."""
        return self.applicantreception_save_action.get_absolute_url()

    @abc.abstractmethod
    def init_provider(self, request, context):
        u"""
        Описание инициализации провайдера данных сетки расписания.

        .. code::

            week = get_weekdays_for_date(context.date)
            date_from, date_to = week[0][1], week[-1][1]

            provider = TimeTableProvider()
            provider.init(
                school_id=core_helpers.get_current_school(request).id,
                date_from=date_from,
                date_to=date_to,
                specialist_id=context.specialist_id)
            provider.load_data()
            return provider
        """

    def init_adapter(self, provider, request, context):
        u"""
        Описание инициализации адаптера данных для грида сетки расписания.
        """
        return providers.TimeTableGridAdapter(provider)

    def delete_applicant_reception(self, request, context):
        u"""Удаление назначения заявителя на прием."""
        try:
            timetablerecord = self.get_obj(context)
        except self.model.DoesNotExist:
            raise m3_actions.ApplicationLogicException(
                self._msg_record_not_found)

        if not timetablerecord.applicantreception_set.exists():
            raise m3_actions.ApplicationLogicException(
                self._msg_no_accepted_records)

        timetablerecord.applicantreception_set.all().delete()

    def save_applicant_reception(self, request, context):
        u"""Назначение на прием."""
        try:
            applicant = models.Applicant.objects.get(id=context.id)
        except models.Applicant.DoesNotExist:
            raise m3_actions.ApplicationLogicException(
                self._msg_applicant_not_found)

        try:
            timetablerecord = self.get_obj(context)
        except self.model.DoesNotExist:
            raise m3_actions.ApplicationLogicException(
                self._msg_record_not_found)

        if context.is_editing:
            # редактирование, нужно удалить предыдущие назначения приема
            timetablerecord.applicantreception_set.all().delete()

        message = accept_reception(applicant, timetablerecord)
        if message:
            # не удалось
            raise m3_actions.ApplicationLogicException(message)

    def extend_menu(self, menu):
        u"""
        Добавление в "Пуск"

        .. code::

            return menu.SubMenu(
                u'Зачисление', menu.Item(self.title, self.window_action))
        """


class RowsAction(objectpack.BaseAction):

    u"""Экшн данных."""

    perm_code = 'view'

    def run(self, request, context):
        u"""
        Отображение данных по строкам.

        :rtype: PreJsonResult
        """
        provider = self.parent.init_provider(request, context)
        adapter = self.parent.init_adapter(provider, request, context)
        # Получение данных для грида: в каждой строке вместо списка словарей -
        # один словарь (ключи не пересекаются)
        result = adapter.get_rows(context.specialist_id)

        return m3_actions.PreJsonResult({'rows': result, 'total': len(result)})


class ApplicantSelectWinAction(objectpack.SelectorWindowAction):

    u"""Окно показа списка посетителей для назначения на прием"""

    perm_code = 'edit'

    multi_select = False

    def configure_action(self, request, context):
        self.callback_url = self.parent.get_applicantreception_save_url()
        self.data_pack = self.parent.applicant_pack

    def configure_window(self, win, request, context):
        action = u'изменить' if context.is_editing else u'добавить'
        win.title = self.parent.applicant_select_win_title_fmt.format(action)
        return win


class ApplicantReceptionSaveAction(objectpack.BaseAction):

    u"""Экшн назначения на прием (изменения назнчения на прием)."""

    perm_code = 'add'

    @commit_on_success
    def run(self, request, context):
        # делегирование паку
        self.parent.save_applicant_reception(request, context)
        return m3_actions.OperationResult()


class ApplicantReceptionDeleteAction(objectpack.BaseAction):

    u"""Экшн удаления назначения заявителя на прием."""

    perm_code = 'delete'

    @commit_on_success
    def run(self, request, context):

        # делегирование паку
        self.parent.delete_applicant_reception(request, context)

        return m3_actions.OperationResult()


class BaseApplicantPack(objectpack.ObjectPack):

    u"""Пак списка посетителей, назначенных и для назначения на прием"""

    model = models.Applicant

    _is_primary_for_model = False

    columns = [{
        'data_index': 'fullname',
        'header': u'Ф.И.О',
        'width': 2
    }]

    MSG_NOT_FOUND = u"Запись не найдена"

    @abc.abstractproperty
    def reception_pack(self):
        u"""
        Пак расписания приема специалистов.

        .. code::

            return ControllerCache.find_pack(ReceptionPack)
        """

    def declare_context(self, action):
        result = super(BaseApplicantPack, self).declare_context(action)
        if action is self.rows_action:
            result[self.reception_pack.id_param_name] = {'type': 'int'}
            result['is_editing'] = {'type': 'boolean', 'default': False}
        return result

    def get_rows_query(self, request, context):
        try:
            record = self.reception_pack.get_obj(context)
        except self.reception_pack.model.DoesNotExist:
            raise m3_actions.ApplicationLogicException(self.MSG_NOT_FOUND)

        query = super(BaseApplicantPack, self).get_rows_query(
            request, context
        )

        if context.is_editing:
            # прием "переназначается на другого заявителя"
            # исключается текущий назначенный
            record = self.reception_pack.get_obj(context)
            try:
                applicant_reception = record.applicantreception_set.all()[0]
            except IndexError:
                # вызвали "Изменить" в ячейке, в кот. еще ничего не назначено
                raise m3_actions.ApplicationLogicException(self.MSG_NOT_FOUND)

            query = query.exclude(id=applicant_reception.applicant_id)

        u"""
        В наследнике предполагается дальнейшая фильтрация посетителей
        с учетом бизнес-логики логики системы

        .. code::

            # в ЭШ посетители - это подавшие документы на зачисление
            query = query.filter(
                school=core_helpers.get_current_school(request),
                # статус "Зарегистрировано"
                status=DECLARATION_STATUS_ACCEPT,
                # дата подачи - сегодня или раньше
                declaration__date__lte=record.begin.date()
            )

            return query.select_related('declaration').order_by(
                'declaration__fullname')
        """
        return query


class BaseSpecialistSelectPack(objectpack.ObjectPack):

    u"""Предок пака выбора специалиста."""

    model = models.Specialist
    _is_primary_for_model = False
    columns = [{
        'data_index': "fullname",
        'header': u'Ф.И.О.',
        'sortable': True,
        'searchable': True,
        'search_fields': ('fullname',),
    }]
    column_name_on_select = 'fullname'
    list_sort_order = ("fullname",)

    def declare_context(self, action):
        result = super(BaseSpecialistSelectPack, self).declare_context(
            action)
        if action is self.rows_action:
            result['date'] = {
                'type': 'date', 'default': datetime.date.today()
            }
        return result

    @abc.abstractmethod
    def _get_timetablecrontab_query(self, request, context):
        u"""
        Получение записей сетки расписания в интервале выбранных дат

        .. code::

            return TimeTableCrontab.objects.filter(
                date_begin__lte=context.date,
            )
        """

    def get_rows_query(self, request, context):
        was_scheduled = self._get_timetablecrontab_query(
            request, context
        ).values_list('specialist')
        query = self.model.objects.filter(id__in=was_scheduled)

        u"""
        В наследнике предполагается фильтрация специалистов
        с учетом бизнес-логики логики системы

        .. code::

            query = query.filter(
                school=core_helpers.get_current_school(request),
                info_date_begin__lte=context.date,
                info_date_end__gte=context.date,
                depersonalized=False,
            )

        В том числе с учетом прав

        .. code::

            # если есть право "просмотр своих" и нет права "просмотр всех"
            if self.parent.has_perm(
                request, 'view_own'
            ) and not self.parent.has_perm(request, 'view_all'):
                query = query.filter(
                    person__userprofile=request.user.get_profile())
        """

        return query
